@@ -79,7 +79,8 @@ variable_body()
 
 keep_system_libs_body()
 {
-	export PKG_CONFIG_PATH="${selfdir}/lib1" LIBRARY_PATH="/test/local/lib"
+	export PKG_CONFIG_PATH="${selfdir}/lib1"
+	eval export "$LIBRARY_PATH_ENV"="/test/local/lib"
 	atf_check \
 		-o inline:"\n" \
 		pkgconf --libs-only-L cflags-libs-only
